#ifndef __rf_object_hh__
#define __rf_object_hh__

#include "rf_common.hh"
#include "rf_types.hh"
#include "pxx_string.hh"

namespace rftype
{

using namespace rfrt ;

class Object
{

  friend class ObjectRef ;

  NO_COPY_CTOR(Object)
  NO_ASSIGN(Object)

private:
  ///
  /// Reference counter
  size_t ref_count ;

public:
  ///
  /// Term class for template instantiations
  static const term_class_t term_class = object_term;
  ///
  /// Exceptions
  class Error {};
  class InvalidParameter : public Error {};
  ///
  /// Constructor
  inline Object () ;
  ///
  /// Virtual destructor
  virtual ~Object () ;
  ///
  /// Virtual operator new(). Object default operator new() is based on
  /// a default refal allocator. However, we assume that operator() may be
  /// overloaded (E.g. for short object we may use some sort of chunk
  /// allocator). But each overloaded operator new() should indirectly imply a
  /// call of default refal allocator, so we can detect in constructor whether
  /// an object was allocated dynamically and setup reference counter
  /// appropriately.
  inline void* operator new (size_t _size) ;
  inline void* operator new (size_t _size, void* _ptr) ;

//protected:
  ///
  /// Virtual operator delete(). It is made protected, because user should not
  /// explicitly delete an object, but an object should be deleted only as a
  /// result of destruction of last reference to it.
  inline void operator delete (void* _ptr) ;

public:
  ///
  /// Get object type. We are not convinced with typeinfo, because we want
  /// to keep control on types ordering.
  virtual unsigned get_type () const = 0 ;
  ///
  /// Compute a hash of an object. We need this to implement table lookup etc.
  virtual uint32_t hash () const ;
  ///
  /// Compare two objects. Refal claims strict ordering, but in real life
  /// we not always have it. Probably this function should throw exception for
  /// non-ordered types. Table implementation, etc. should not rely on this
  /// function.
  /// This function should compare object types first.
  virtual inline int compare (Object const& _obj) const ;
  ///
  /// Check whether two objects are equal.
  virtual bool operator == (Object const& _obj) const ;
  ///
  /// Check whether two objects are not equal. Default implementation implies
  /// equality check.
  virtual inline bool operator != (Object const& _obj) const ;

  virtual inline pxx::WString to_string () const;

  virtual inline size_t get_name (wchar_t const** _name) const;
};

}

#endif // __rf_object_hh__
