//-----------------------------------------------------------------------------
/// @file rf_term.hh
///
/// Refal+ term interface
//
//  $Source$
//  $Revision: 2187 $
//  $Date: 2006-12-15 03:17:56 +0300 (Пт, 15 дек 2006) $
//-----------------------------------------------------------------------------

#ifndef __rf_term_hh__
#define __rf_term_hh__

#include "rf_memory_chunk.hh"
#include "rf_common.hh"
#include "rf_types.hh"
#include "pxx_string.hh"

///
/// A namespace for Refal runtime
namespace rfrt
{

template <class C1, class C2>
class OR;

//
// In previous runtime version we keeped the order of memory block in each
// reference term. This was needed to compute a pointer of a memory block.
// Now an allocator is able to get this pointer for us, so we don't need all
// stuff related to this. However, we still keep old code, because in a future
// we may return to the old scheme if it can lead to faster program execution.
#if defined(RF_STORE_CHUNK_ORDER_IN_TERM)

#if UINTPTR_MAX == UINT32_MAX
#define ORDER_SHIFT 3
#define REF_BIT (uintptr_t(0x04))
#define PTR_MASK (~(uintptr_t(0x07)))
#define ORDER1_MASK (uintptr_t(0x03))
#define ORDER2_MASK (uintptr_t(0x07))
#elif UINTPTR_MAX == UINT64_MAX
#define ORDER_SHIFT 4
#define REF_BIT (uintptr_t(0x08))
#define PTR_MASK (~(uintptr_t(0x0F)))
#define ORDER1_MASK (uintptr_t(0x03))
#define ORDER2_MASK (uintptr_t(0x0F))
#else
#error "Unsupported pointer size"
#endif
//#define COMPOUND_BIT (uintptr_t(0x8000))
//#define CHAR_SHIFT ((sizeof(uintptr_t) - sizeof (uint16_t)) * 8)
#define ORDER1(x) ((uintptr_t)((x) >> ORDER_SHIFT))
#define ORDER2(x) ((uintptr_t)((x) & ORDER2_MASK))
#define ORDER(x,y) (((x) << ORDER_SHIFT) | (y))

#endif // RF_STORE_CHUNK_ORDER_IN_TERM

class Expr ;

//-----------------------------------------------------------------------------
/// Refal term. This is one of two basic runtime classes, which describes
/// Refal expression elements.
///
/// FIXME: To follow all invariants it is necessary to define copy ctor and
/// assignment operator, which should increment memory block reference counter
/// in a case of reference term. Destructor should decrement this reference
/// counter appropriately. This is not done for optimization purposes.
//-----------------------------------------------------------------------------
class Term
{
///
/// We can access all term fields from expression class and some methods
friend class Expr ;
friend inline Expr operator + (Term const&, Expr const&) ;
friend inline Expr operator + (Term const&, Term const&) ;

protected:
  ///
  /// The first term field containing symbol type
  uintptr_t data1 ;
  ///
  /// The second term field containing actual data (for short type) or pointer
  /// to object (for object extensions).

  union {
    uintptr_t uint_data2 ;
    intptr_t int_data2 ;
    void* ptr_data2 ;
    void (*fptr_data2) () ;
  };

  ///
  /// Constructor for empty term with given class and type
  inline Term (unsigned _type) ;
  ///
  /// Constructor for pointer term of given type (to use for static terms)
  inline Term (unsigned _type, void* _ptr) ;
  ///
  /// Create expression with specified length
  static inline Expr create_expr (size_t _len, int _align = 0) ;
  ///
  /// Create flat expression with specified length
  static inline Expr create_flat_expr (size_t _len, int _align = 0) ;
private:
  ///
  /// Constructor for a character term. All constructors are private because
  /// they should be called only from expression implementation.
  /// FIXME: Originally it was planned that all strings in Refal will be
  /// unicode-based (there will be no difference in expression size, because
  /// a term can contain either usual character or wide character). However it
  /// was decided later to have both character types, so probably instead of
  /// this we should have Term (char _c) and Term (wchar_t _c).
//  inline Term (uint16_t _c) ;

  inline Term (Term const& _t) ;
  inline Term& operator = (Term const& _t) ;

public:
  ///
  /// Destructor.
  inline ~Term () ;
  ///
  /// Check whether two terms are equal
  inline bool operator == (Term const& _t) const ;
  ///
  /// Check whether two terms are not equal
  inline bool operator != (Term const& _t) const ;
  ///
  /// Check whether two term sequences are equal
  static inline bool eq (
    Term const* _f1, Term const* _l1, Term const* _f2, Term const* _l2
  ) ;
  static inline bool eq (
    Term const* _p1, Term const* _p2, uintptr_t _len
  ) ;
  ///
  /// Check whether a term is a symbol
  inline bool is_sym () const ;
  ///
  /// Check whether a term is a reference term
  inline bool is_ref () const ;
  ///
  /// Check whether a term is a border symbol
  inline bool is_border () const ;
  ///
  /// Get a type of "short" symbol.
private:
  ///
  /// Return a type of "short" symbol or type_ref or type_object.
  inline unsigned get_short_type () const ;
public:
  ///
  /// Set "short" type of the term.
  inline void set_type (unsigned _type) ;
  ///
  /// Get concrete type of the term.
  inline unsigned get_type () const ;
  ///
  /// Cast to the concrete term type.  If not success return null.
  template <class C>
  inline C* cast_to (unsigned _type) ;

  template <class C>
  inline C* cast_to () ;
  ///
  /// Get a reference to underlying object.  The object must have type C.
  template <class C>
  inline C const& get_object () const ;
  ///
  /// Get a pointer to underlying object for changing it.  The object must have
  /// type C and the term must be changable (i.e. to be reference term).
  template <class C>
  inline C* get_obj_ptr () const ;
  ///
  /// Check that underlying object has type C.
  template <class C>
  inline bool is_instance_of () const ;
  ///
  /// Check that underlying object satisfies the TypeCond::check() condition.
  template <class TypeCond>
  inline bool check_type () const ;
  ///
  /// Destruct underlying object, then construct a new one in its place.  The
  /// object must have type C and the term must be changable (i.e. to be
  /// reference term).  Old object is destroyed at first, so it can't be used
  /// in construction of a new one.
  template <class ObjClass>
  inline void replace_object () ;

  template <class ObjClass, class CppType1>
  inline void replace_object (CppType1 const& _arg1) ;

  template <class ObjClass, class CppType1, class CppType2>
  inline void replace_object (CppType1 const& _arg1, CppType2 const& _arg2) ;
  
  template <class ObjClass, class CppType1, class CppType2, class CppType3>
  inline void replace_object (
      CppType1 const& _arg1, CppType2 const& _arg2, CppType3 const& _arg3) ;
  
  inline operator pxx::WString () const ;
  ///
  /// If term has a name then make _np point to it and return its length.
  /// Else return (size_t)-1.
  inline size_t get_name (wchar_t const** _np) const ;

  static inline int compare (Term const& _t1, Term const& _t2) ;
  static inline int compare (
    Term const* _f1, Term const* _l1, Term const* _f2, Term const* _l2
  ) ;

  template <class C1, class C2>
  class Convert
  {
  public:
    static C2 convert (C1 const& _obj)
    {
      return _obj;
    }
  };

  template <class C1, class C2>
  class ConvertObject
  {
  public:
    static bool convert (Term const& _t, C2* _p)
    {
      if (!_t.is_instance_of<C1>())
        return false;
      *_p = Convert<C1, C2>::convert(_t.get_object<C1>());
      return true;
    }
  };

  template <class C11, class C12, class C2>
  class ConvertObject<OR<C11, C12>, C2>
  {
  public:
    static bool convert (Term const& _t, C2* _p)
    {
      if (ConvertObject<C11, C2>::convert(_t, _p))
        return true;
      return ConvertObject<C12, C2>::convert(_t, _p);
    }
  };

  template <class C1, class C2, class Cond>
  class CheckConvert
  {
  public:
    static int convert (Term const& _t, C2* _p)
    {
      if (!_t.is_instance_of<C1>())
        return -1;
      C1 obj = _t.get_object<C1>();
      if (!Cond::check(obj))
        return 0;
      *_p = Convert<C1, C2>::convert(obj);
      return 1;
    }
  };

  template <class C11, class C12, class C2, class Cond>
  class CheckConvert<OR<C11, C12>, C2, Cond>
  {
  public:
    static int convert (Term const& _t, C2* _p)
    {
      int c = CheckConvert<C11, C2, Cond>::convert(_t, _p);
      if (c == -1)
        return CheckConvert<C12, C2, Cond>::convert(_t, _p);
      return c;
    }
  };

#if defined(RF_STORE_CHUNK_ORDER_IN_TERM)
  inline unsigned get_order () const
  {
    return ORDER(data1 & ORDER1_MASK, data2 & ORDER2_MASK);
  }
#endif

};

}

#endif // __rf_term_hh__
